#pragma once 

#include "texture_atlas.hpp"

#include "stb_image.h"

#include <gz-util/string/utility.hpp>
#define LOG_SUBLOGS
#include <gz-util/log.hpp>
#include <glm/fwd.hpp>

#include <unordered_map>

namespace gz::vlk {
    // TODO use this texture handle
    using Texture = uint32_t;
    using TextureAtlasIndex = uint32_t;

    struct TextureInfo {
        TextureAtlasIndex atlas;
        glm::vec2 texCoordTopLeft;
        glm::vec2 texCoordBottomRight;
        std::string toString() const;
    };
    /// Defined in vulkan_instance.hpp
    class VulkanInstance;
    /**
     * @details
     *  Manages multiple @ref TextureAtlas "texture atlases", which each hold one image that contains muiltiple textures
     *  @section tm_descriptors
     *   The TextureManager has provides a descriptor set for accessing the texture atlases:
     *   Bindings:
     *   -0. Array of combined image samplers (DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER) of length (getAtlasCount())
     */
    class TextureManager {
        public:
            /**
             * @brief Create a texture manager
             * @details 
             *  -# @ref VulkanInstance::registerCleanupCallback "register" @ref cleanup() "cleanup callback"
             *  -# @ref createDescriptor  "create descriptor set layout, pool and set
             */
            TextureManager(VulkanInstance& instance);
            /**
             * @brief Get a texture handle for a texture. Loads the texture from file if not already loaded
             */
            Texture getTexture(const std::string& textureName);

            /**
             * @brief Set the texture coordiantes for the texture
             * @details
             *  @param texCoords 
             *   The texture coordinates to set. texCoords should be from (0,0) to (1,1) and will be 
             *   set to the according texture coordinate of texture.
             */
            void setTexCoords(const Texture& texture, glm::vec2& texCoords);
            /* void getTexCoords(const std::string& textureName, glm::vec2& texCoords); */
			const vk::DescriptorSet& getDescriptorSet() const { return descriptorSet; }
			const vk::DescriptorSetLayout& getDescriptorSetLayout() const { return descriptorSetLayout; }
            /**
             * @brief Get the atlas that texture is in
             * @throws InvalidArgument if the texture is not present
             */
            const TextureAtlas& getTextureAtlas(const Texture& texture) const;
            /* const TextureAtlas& getTextureAtlas(const std::string& texture) const; */
            /**
             * @brief Get the index of the atlas that texture is in
             * @throws InvalidArgument if the texture is not present
             */
            TextureAtlasIndex getTextureAtlasIndex(const Texture& texture) const;
            /* TextureAtlasIndex getTextureAtlasIndex(const std::string& texture) const; */
            /**
             * @brief Get the number of atlases. 
             * @details Needed for specialization constant in shaders using sampler arrays
             */
            uint32_t getAtlasCount() const { return static_cast<uint32_t>(atlases.size()); }

            /**
             * @brief Add a new atlas and a special "atlas_X" texture (where X is the returned index)
             */
            TextureAtlasIndex addAtlas(uint16_t slotWidth, uint16_t slotHeight, uint16_t slotCountX, uint16_t slotCountY, vk::Format imageFormat = vk::Format::eR8G8B8A8Srgb);


        private:
            void cleanup();
            /**
             * @brief Check if texture handle is valid
             * @throws InvalidArgument if texture is not valid
             */
            void isValidTexture(const Texture& texture) const;
            Texture loadTextureFromFile(const std::string& textureName);
            std::vector<TextureAtlas> atlases;
            /**
             * @brief Contains textureName : Texture pairs
             */
            util::unordered_string_map<Texture> textures;
            std::vector<TextureInfo> textureInfos;
            VulkanInstance& vk;
            Log tLog;
        /**
         * @name Create desciptors
         * @see tm_descriptors
         */
            /// @{
            vk::DescriptorSetLayout descriptorSetLayout;
            vk::DescriptorPool descriptorPool;
            vk::DescriptorSet descriptorSet;
            void createDescriptorResources();
            /// @}


    };
}  // namespace gz::vlk
